//
// RTL Grid system
// --------------------------------------------------

// Columns
//
// Common styles for small and large grid columns

.make-rtl-grid-columns(@rtl-prefix);


// Extra small grid
//
// Columns, offsets, pushes, and pulls for extra small devices like
// smartphones.

.make-rtl-grid(xs, @rtl-prefix);


// Small grid
//
// Columns, offsets, pushes, and pulls for the small device range, from phones
// to tablets.

@media (min-width: @screen-sm-min) {
  .make-rtl-grid(sm, @rtl-prefix);
}


// Medium grid
//
// Columns, offsets, pushes, and pulls for the desktop device range.

@media (min-width: @screen-md-min) {
  .make-rtl-grid(md, @rtl-prefix);
}


// Large grid
//
// Columns, offsets, pushes, and pulls for the large desktop device range.

@media (min-width: @screen-lg-min) {
  .make-rtl-grid(lg, @rtl-prefix);
}
