/* This file is part of Tryton.  The COPYRIGHT file at the top level of
   this repository contains the full copyright notices and license terms. */
@brand-primary: #267f82;
@brand-success: #32ad52;
@brand-info: #71bdc1;
@brand-warning: #e78e42;
@brand-danger: #e74842;

@state-selected-text: lighten(@brand-primary, 25%);
@state-selected-bg: lighten(@brand-primary, 50%);
@state-selected-border: darken(spin(@state-selected-bg, -10), 5%);

@input-border-focus: lighten(@brand-primary, 25%);

@navbar-inverse-bg: @brand-primary;

@navbar-inverse-link-color: lighten(@brand-primary, 30%);

@navbar-inverse-toggle-hover-bg: rgba(0,0,0,0.5);
@navbar-inverse-toggle-icon-bar-bg: rgba(0,0,0,0.5);

@nav-tabs-justified-link-border-color: @nav-tabs-border-color;

@panel-default-text: #fff;
@panel-default-heading-bg: @brand-primary;

@icon-font-path: "../bower_components/bootstrap/fonts/";

@font-size-small: @font-size-base;
